/*

  Charles E V Pegge  
  12:18 08/11/2013

  PROPERTIES:
  ===========

  * Standard header base used for all classes of object

  * Expandable Array of Objects as single entity

  * Array buffer can be lerger than contents, to minimise memory reallocations

  * Array buffer is terminated by 2 extra null bytes (supporting wide character arrays)


  STRUCTURE:
  ==========

  BASE-HEADER
  ADDITIONAL MEMBERS
  OBJECT ARRAY BUFFER

*/


  #define FreeSpace   free
  #define FreeObject  free
  #define ObjectOffset(A) A + A->offset
  #define ObjectHeaderBytes(A) A->offset
  #define ObjectAllocBytes(A) A->offset + A->nbytes + 2


  typedef struct ClassTableStruct ClassTable,*Methods;

  class ClassStruct
  begin
  Methods f;  // pointer to methods/function table
  int type;   // dynamic type code
  int offset; // size of this header (jump to data)
  int nbytes; // buffer size for body of data
  int count;  // number of elements
  int size;   // size of each element
  end
  Class,*Object;


  //BYTES

  function int   ByteLen(char *s);
  function int   WordLen(char *s);
  function ref   NewSpace        (int count);
  sub            CopyBytes       (ref tv,ref sv,int count);
  sub            NullTerminate   (char *s);
  function int   FillBytes       (char *s, int le, char c);
  function ref   AdjustBuffer    (ref s,int m, int n);

  //GENERIC OBJECTS
  
  function ref   Transfer        (ref *r, ref s);
  function ref   Stretch         (ref *pt,int n);
  function ref   Shrink          (ref *pt, int n);
  function ref   Truncate        (ref *pt, int n);
  sub            SetObjectHeader (Object this,ref f,int le, int wi,int so);
  function ref   NewObject       (ref f, int nc, int wi, int so);


  dim Methods pMethods=0;


  function int ByteLen(char*s)
  begin
    int i=0;
    while (s[i] !=0 ) i++;
    return i;
  end


  function int WordLen(char *s)
  begin
    short *w=(short*) s;
    int i=0;
    while (w[i] !=0 ) i++;
    return i;
  end


  function ref NewSpace(int count)
  begin
    char *v =  malloc(count);
    int i;
    for (i=0; to i<count; step incr i) v[i]=0; // set chars to null
    return v;
  end


  sub CopyBytes(ref tv,ref sv,int count)
  begin
    char *s=sv;
    char *t=tv;
    int i;
    for (i=0; to i<count; step incr i)
    begin
      *t=*s;
   incr t;
      incr s;
    end
  end


  sub NullTerminate(char *s)
  begin
    s[0]=0;
    s[1]=0;
  end


  function int FillBytes(char *s, int le, char c)
  begin
    int i;
    for (i=0; to i<le; incr i)
    begin
      *s=c;
      incr s;
    end
    return 1;
  end


  function ref AdjustBuffer(ref s,int m, int n)
  begin
    char *t=malloc(n);
    if (m>n)  then m=n; //clamp m
    if (s==0) then m=0;
    if (m)    then CopyBytes(t,s,m);
    FillBytes(t+m,n-m,0);
    return t;
  end


  function ref Transfer(ref *r, ref s)
  begin
    if (*r==s) return s;   // no transfer
    if (*r) then free(*r); // free old
    *r=s;                  // ref new
    return s;              // return new
  end


  function ref Stretch(ref *pt,int n)
  begin
    Object t=*pt;
    if (t==0) then return 0;
    int mm=t->count * t->size + t->offset;
    int nn=n * t->size + t->offset;
    if (nn > t->nbytes) then AdjustBuffer(t, mm, nn+2);
    Transfer(pt,t);
  end


  function ref Shrink(ref *pt, int n)
  begin
    Object t=*pt;
    if (t==0) then return 0;
    int mm=t->count * t->size + t->offset;
    int nn=n * t->size + t->offset;
    if (nn<mm) then nn=mm; //clamp minimum nn
    if (nn < t->nbytes) then AdjustBuffer(t, mm, nn+2);
    Transfer(pt, t);
  end


  function ref Truncate(ref *pt, int n)
  begin
    Object t=*pt;
    if (t==0) then return 0;
    int mm=t->offset;
    int nn=n * t->size + t->offset;
    if (nn<mm) then nn=mm; //clamp minimum nn
    if (nn < t->nbytes) then AdjustBuffer(t, mm, nn+2);
    Transfer(pt, t);
  end


  sub SetObjectHeader(Object this,ref f,int le, int wi,int so)
  begin
    this->f      = f;
    this->offset = so;
    this->nbytes = le*wi;
    this->count  = le;
    this->size  = wi;
  end


  function ref NewObject(ref f, int nc, int wi, int so)
  begin
    Object this = NewSpace(nc*wi+so+2);
    SetObjectHeader(this, f, nc, wi,so);
  end




