
/*

  LEXING
  ======

*/

  /* Word -> x flags */  

  #define IsNull    1
  #define LineBreak 2
  #define IsWord    4
  #define IsNumber  8
  #define IsAlpha   16
  #define IsQuote   32
  #define IsSymbol  64
  #define Comment  128


  class WordClassStruct
  begin
//BASE MEMBERS
  StringMethods f;    // pointer to methods table
  int type;           // dynamic type code
  int offset;         // size of this header
  int nbytes;         // buffer size for body of data
  int count;          // number of elements
  int size ;          // size of each element
//ADDITIONAL MEMBERS
  StringObject text;  // string buffer
  char * bp;           // start of word
  char * ep;           // end boundary of word
  char * et;           // end boundary of text
  char * next;         // next word
  int   xx;           // status bits
  end
  WordClass, *WordObject;


  function char *WordInfo(WordObject*pw)
  begin
    WordObject w=*pw;
    char *s=(ref) w + w->offset;
  //printf("TEXT:       %s\n",w->bp);
    printf("WORD:       %s\n",s);
    printf("Class Size: %i\n",w->offset);
    printf("Char Width: %i\n",w->size);
    printf("Word Len:   %i\n",w->count);
    printf("Word Start: %i\n",w->bp);
    printf("Word End:   %i\n",w->ep);
    printf("Text End:   %i\n",w->et);
  end

  function char *WordShow(WordObject *pw)
  begin
    WordObject w=*pw;
    char *s=(ref) w + w->offset;
    printf("%s\n",s);
  end

  function WordObject NewWord(StringObject *pText, int n)
  begin
    StringObject text=*pText;
    if (n<=0) n=64;
    WordObject w=NewSpace(sizeof(WordClass)+n*text->size);
    w->nbytes=n * text->size;
    w->offset=sizeof(WordClass);
    w->count=n;
    w->size=text->size;
    w->text=text;
    w->bp  =(ref) text+text->offset;
    w->next=(ref) text+text->offset;
    w->et  =(ref) text + text->offset + (text->count * text->size);
    return w;
  end


  method WordObject WordFree(WordObject *pthis)
  begin
    WordObject this= *pthis;
    if (this==0) FreeSpace((ref)this);
    *pthis=0;
    return 0;
  end


  function WordObject SkipSpace(WordObject w)
  begin
    int ex=0;
    char *n = w->et;
    while (w->next < n)
    begin
      switch (*w->next)
      begin
        case 0  ... 32  : {break;}
        default         : {ex=1; w->xx |= IsWord; break;}
      end
      if (ex) then break;
      ++w->next;
    end
    return w;
  end


  function WordObject SkipSpaceL(WordObject w)
  begin
    int ex=0;
    char *n = w->et;
    while (w->next < n)
    begin
      switch (*w->next)
      begin
        case 1  ...  9  : {break;}
        case 10         : {ex=1; w->xx |=LineBreak; break;}
        case 11         : {break;}
        case 12         : {ex=1; w->xx |=LineBreak; break;}
        case 13         : {ex=1; w->xx |=LineBreak; break;}
        case 14 ... 32  : {break;}
        default         : {ex=1; w->xx |=IsWord; break;}
      end
      if (ex) then break;
      ++w->next;
    end
    return w;
  end


  function char *NextSpace(WordObject w)
  begin
    int ex=0;
    char *n = w->et;
    while (w->next < n)
    begin
      switch (*w->next)
      begin
        case  0 ... 32 : {ex=1; break;}
      end
      if (ex) then break;
      ++ w->next;
    end
  end


  function char *EndLine(WordObject w)
  begin
    int ex=0;
    char *n = w->et;
    while (w->next < n)
    begin
      switch (*w->next)
      begin
        case 10 : {ex=1; break;}
        case 13 : {ex=1; break;}
      end
      if (ex) then break;
      ++ w->next;
    end
  end


  function char *SkipQuoteMarks(WordObject w)
  begin
    char c=*w->next;
    w->next++;
    char *n = w->et -1;
    while ((w->next < n)and(c != *w->next))
    begin
      ++w->next;
    end
    return w->next+1;
  end


  function WordObject NextBasicWord(WordObject *pw)
  begin
    WordObject w=*pw;
    if (w==0) then return 0;
    SkipSpace(w);
    w->xx = 0;
    int ex=0;
    w->bp=w->next;
    switch (*w->next)
    begin
      case '0' ... '9': {w->xx |=IsNumber; break;}
      case 'A' ... 'Z': {w->xx |=IsAlpha;  break;}
      case 'a' ... 'z': {w->xx |=IsAlpha;  break;}
      case '_'        : {w->xx |=IsAlpha;  break;}
      case 0          : {                  break;}
      case 34         : {w->xx |=IsQuote;  break;}
      case 96         : {w->xx |=IsQuote;  break;}
      default         : {w->xx |=IsSymbol; break;}
    end
    char *n = w->et;
    while (w->next < n)
    begin
      switch (*w->next)
      begin
        case '0' ... '9': {break;}
        case 'A' ... 'Z': {break;}
        case 'a' ... 'z': {break;}
        case '_'        : {break;}
        case 13         : {ex=1; break;}
        case 10         : {ex=1; break;}
        case 34         : {SkipQuoteMarks(w); ex=1; break;}
        case 96         : {SkipQuoteMarks(w); ex=1; break;}
        default         : // symbols 
        begin
          if(w->bp == w->next) ++w->next;
          ex=1;
          break;
        end
      end
      if (ex) then break;
      ++ w->next;
    end;
    w->ep = w->next;
    w->count = w->ep - w->bp;
    int le=w->count * w->size;
    if (le > w->nbytes) //stretch word buffer
    begin
      char *t = malloc(w->offset + w->nbytes + 34);
      CopyBytes(t, w, w->offset);
      free(w);
      w = (ref) t;
    end
    char *k = (ref) w + w->offset;
    CopyBytes(k, w->bp, le);
    NullTerminate(k+le);
    SkipSpaceL(w);
    return Transfer((ref*)pw, (ref)w);
  end

//itr case conversion
